#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "log.h"
#include "common/xmalloc.h"

#define REPAIRING_DEVICE_FILE	"/sys/kernel/repairing_device"
#define ASCII_SIZE	128

static void *root_bucket[ASCII_SIZE];

static int insert_devname_internal(void *bucket[], const char *name)
{
	void *new = NULL;
	int idx = *name;

	if (bucket[idx] != NULL)
		return insert_devname_internal(bucket[idx], name+1);
	else if (idx == '\0') {
		new = xmalloc(sizeof(void *));
		if (!new) {
			pr_perror("alloc devname failed\n");
			return -1;
		}
		bucket[idx] = new;
		return 0;
	} else {
		new = xmalloc(sizeof(void *) * ASCII_SIZE);
		if (!new) {
			pr_perror("alloc devname failed\n");
			return -1;
		}
		memset(new, 0, sizeof(void *) * ASCII_SIZE);
		bucket[idx] = new;
		return insert_devname_internal(bucket[idx], name+1);
	}
}

int insert_devname(const char *devname)
{
	if (devname == NULL || *devname == '\0') // ignore
		return 0;

	pr_debug("insert device '%s'\n", devname);
	return insert_devname_internal(root_bucket, devname);
}

int parse_devname(void)
{
	int retval = -1;
	char *line = NULL;
	size_t len = 0;
	ssize_t nread = 0;
	FILE *fp = NULL;

	fp = fopen(REPAIRING_DEVICE_FILE, "r");
	if (fp == NULL) {
		pr_info("Unable to open %s, downgrade to use internal whitelist\n",
				REPAIRING_DEVICE_FILE);
		return 0;
	}

	while ((nread = getline(&line, &len, fp)) != -1) {
		if (nread <= 1) // ignore empty string
			continue;

		line[nread-1] = '\0'; // drop '\n'
		retval = insert_devname(line);
		if (retval != 0)
			goto out;
	}
	retval = 0;

out:
	free(line);
	fclose(fp);
	return retval;
}

static const char *steal_devname(const char *name, ssize_t len)
{
	ssize_t off = len;

	for (off -= 1; off > 0; off--) {
		if (name[off] == '/')
			break;
	}

	return name + off + 1;
}

static bool find_devname_internal(void *bucket[], const char *name)
{
	int idx = *name;

	if (*name == '\0' && bucket[idx] != NULL)
		return true;
	else if (bucket[idx] == NULL)
		return false;
	else {
		return find_devname_internal(bucket[idx], name+1);
	}
}

bool find_devname(const char *name)
{
	const char *devname;
	size_t len = 0;
	bool found = false;

	if (name == NULL)
		return false;
	else if ((len = strlen(name)) == 0)
		return false;

	devname = steal_devname(name, len);
	found = find_devname_internal(root_bucket, devname);

	pr_debug("device '%s' (original name '%s') %s found in %s\n",
			devname, name, found ? "is" : "isn't", REPAIRING_DEVICE_FILE);

	/* Compatible with the old version, there are still `strstr` branch in the following */
	found |= (strstr(name, "uverbs") != NULL
			|| strstr(name, "rdma_cm") != NULL
			|| strstr(name, "umad") != NULL);

	return found;
}
