#include <sys/ioctl.h>

#include "imgset.h"
#include "pstree.h"
#include "files-chr.h"
#include "log.h"

#include "protobuf.h"
#include "rst-malloc.h"

static unsigned hisi_sec_fds_n;

/* Checks if file descriptor @lfd is infinibandevent */
int is_infiniband_link(char *link)
{
	return is_anon_link_type(link, "[infinibandevent]");
}

static int chrfile_open(struct file_desc *d, int *new_fd)
{
	int fd, mntns_root;
	int ret = -1;
	struct chrfile_info *ci;

	ci = container_of(d, struct chrfile_info, d);

	pr_info("charfile: Opening %s (repair %d index %d)\n",
			ci->path, ci->cfe->repair, ci->cfe->index);

	if (ci->cfe->repair)
		ci->cfe->flags |= O_REPAIR;

	mntns_root = open_pid_proc(getpid());
	fd = openat(mntns_root, ci->path, ci->cfe->flags);
	if (fd < 0){
		pr_err("open chr file failed\n");
		return -1;
	}

	*new_fd = fd;
	ret = 0;

	return ret;
}

static struct file_desc_ops chrfile_desc_ops = {
	.type = FD_TYPES__CHR,
	.open = chrfile_open,
};

static int collect_one_chrfile(void *o, ProtobufCMessage *base, struct cr_img *i)
{
	struct chrfile_info *ci = o;
	static char dot[] = ".";

	ci->cfe = pb_msg(base, ChrfileEntry);
	if (ci->cfe->name[1] == '\0')
		ci->path = dot;
	else
		ci->path = ci->cfe->name;

	/* collect `/dev/hisi_sec2*` fds */
	if (strstr(ci->path, HISI_SEC_DEV) != NULL)
		hisi_sec_fds_n += 1;

	pr_info("Collected chr file: %#x, name: %s\n", ci->cfe->id, ci->path);
	return file_desc_add(&ci->d, ci->cfe->id, &chrfile_desc_ops);
}

struct collect_image_info chrfile_cinfo = {
	.fd_type = CR_FD_CHRFILE,
	.pb_type = PB_CHRFILE,
	.priv_size = sizeof(struct chrfile_info),
	.collect = collect_one_chrfile,
};

static int handle_hisi_vma(struct list_head *fds, struct vma_area *vma);
int collect_chr_map(struct pstree_item *me, struct vma_area *vma)
{
	struct list_head *list = &rsti(me)->fds;
	struct fdinfo_list_entry *fle, *tmp;
	struct chrfile_info *ci;
	bool exist_fd;

	if (strstr(vma->e->name, HISI_SEC_DEV) != NULL) {
		if (handle_hisi_vma(list, vma) != 0) {
			return -1;
		} else
			goto out;
	}

	list_for_each_entry_safe(fle, tmp, list, ps_list) {
		struct file_desc *d = fle->desc;

		if (d->ops->type != FD_TYPES__CHR)
			continue;

		ci = container_of(d, struct chrfile_info, d);
		if (!strcmp(vma->e->name, ci->path)) {
			vma->vmfd = d;
			vma->e->fd = fle->fe->fd;
			exist_fd = true;
			break;
		}
	}

	if (!exist_fd)
		return -EEXIST;

out:
	pr_info("  `- find fd %ld for dev %s at this vma\n", vma->e->fd, vma->e->name);

	return 0;
}

#define MAX_HISI_SEC_SIZE	3 /* one physical device expose three char dev */
static struct hlist_head hisi_sec_fds_hash[MAX_HISI_SEC_SIZE];

static int collect_hisi_sec_fds(struct list_head *list)
{
	struct fdinfo_list_entry *fle, *tmp;
	struct chrfile_info *ci;
	struct file_desc *d;
	struct hisi_sec_desc *desc;
	int idx;
	int nr = 0;

	for (idx = 0; idx < MAX_HISI_SEC_SIZE; idx++)
		INIT_HLIST_HEAD(&hisi_sec_fds_hash[idx]);

	list_for_each_entry_safe(fle, tmp, list, ps_list) {
		d = fle->desc;

		if (d->ops->type != FD_TYPES__CHR)
			continue;

		ci = container_of(d, struct chrfile_info, d);

		if (strstr(ci->path, HISI_SEC_DEV) != NULL) {
			desc = shmalloc(sizeof(*desc));
			if (desc == NULL)
				return -ENOMEM;

			desc->name = ci->path;
			desc->fd = fle->fe->fd;
			desc->mmio = desc->dus = 0;

			idx = (ci->path[strlen(ci->path)-1] - '0') % MAX_HISI_SEC_SIZE;
			hlist_add_head(&desc->hash, &hisi_sec_fds_hash[idx]);

			nr += 1;
		}
	}

	return nr;
}

static long delivery_hisi_sec_fd(struct list_head *fds, struct vma_area *vma)
{
	extern unsigned hisi_sec_fds_n; /* defined in criu/files.c */
	static bool initialized  = false;
	struct hisi_sec_desc *desc;
	int fd = -1, idx;

	if (!initialized) {
		int nr;

		pr_info("find %d fds for hisi_sec char device\n", hisi_sec_fds_n);

		nr = collect_hisi_sec_fds(fds);
		if (nr != hisi_sec_fds_n) {
			pr_err("Collected fds(%d) aren't equal opened(%d)\n",
			       nr, hisi_sec_fds_n);
			return -1;
		}

		initialized = true;
	} else if (vma->e->pgoff != HISI_SEC_MMIO && vma->e->pgoff != HISI_SEC_DUS) {
		/* It's impossible value for fd, just as a tag to show it's a
		 * vma by `mprotect` syscall.
		 */
		return LONG_MAX;
	}

	idx = (vma->e->name[strlen(vma->e->name)-1] - '0') % MAX_HISI_SEC_SIZE;
	hlist_for_each_entry(desc, &hisi_sec_fds_hash[idx], hash) {
		if (strcmp(desc->name, vma->e->name) != 0)
			continue;

		if (vma->e->pgoff == HISI_SEC_MMIO && !desc->mmio) {
			fd = desc->fd;
			desc->mmio = true;
			break;
		} else if (vma->e->pgoff == HISI_SEC_DUS && !desc->dus) {
			fd = desc->fd;
			desc->dus = true;
			break;
		}
	}

	return fd;
}

static int handle_hisi_vma(struct list_head *fds, struct vma_area *vma)
{
	long fd = delivery_hisi_sec_fd(fds, vma);

	if (fd < 0) {
		pr_err("find fd for char dev vma pgoff %lx named %s failed.\n",
		       vma->e->pgoff, vma->e->name);
		return -1;
	}

	vma->e->fd = fd;

	return 0;
}

static void pr_info_infiniband(char *action, InfinibandEntry *infiniband)
{
	pr_info("%sinfiniband: id %#08x\n", action, infiniband->id);
}

static int dump_one_infiniband(int lfd, u32 id, const struct fd_parms *p)
{
	FileEntry fe = FILE_ENTRY__INIT;
	InfinibandEntry infiniband = INFINIBAND_ENTRY__INIT;

	infiniband.id = id;

	fe.type  = FD_TYPES__INFINIBAND;
	fe.id = infiniband.id;
	fe.infiniband = &infiniband;

	pr_info_infiniband("Dumping ", &infiniband);

	return pb_write_one(img_from_set(glob_imgset, CR_FD_FILES), &fe, PB_FILE);
}

const struct fdtype_ops infiniband_dump_ops = {
	.type	= FD_TYPES__INFINIBAND,
	.dump	= dump_one_infiniband,
};

static int infiniband_open(struct file_desc *d, int *new_fd) {
	/*
	 * `*new_fd == -1` at this time, it means this open operation shouldn't
	 * be served out, which is why this function does nothing here.
	 */
	return 0;
};

static struct file_desc_ops infiniband_desc_ops = {
	.type	= FD_TYPES__INFINIBAND,
	.open	= infiniband_open,
};

static int collect_one_infiniband(void *o, ProtobufCMessage *base, struct cr_img *i)
{
	struct infiniband_file_info *info = o;

	info->infiniband = pb_msg(base, InfinibandEntry);
	pr_info_infiniband("Collected ", info->infiniband);

	/* add the fd to `file_desc_hash` list to prevent from NULL pointer */
	return file_desc_add(&info->d, info->infiniband->id, &infiniband_desc_ops);
}

struct collect_image_info infiniband_cinfo = {
	.fd_type = CR_FD_INFINIBAND,
	.pb_type = PB_INFINIBAND,
	.priv_size = sizeof(struct infiniband_file_info),
	.collect = collect_one_infiniband,
};
