#include <sys/socket.h>
#include <sys/un.h>
#include <sys/mount.h>
#include <unistd.h>
#include <errno.h>
#include <stdbool.h>

#include "int.h"
#include "types.h"
#include "common/compiler.h"
#include "fcntl.h"
#include "log.h"
#include "util-pie.h"
#include "notifier.h"

#ifdef CR_NOGLIBC
#include <compel/plugins/std/syscall.h>
#define __sys(foo) sys_##foo
#else
#define __sys(foo) foo
#endif

#ifdef CR_NOGLIBC
#define __pr_perror(fmt, ...) pr_err(fmt "\n", ##__VA_ARGS__)
#else
#define __pr_perror(fmt, ...) pr_perror(fmt, ##__VA_ARGS__)
#endif

int open_detach_mount(char *dir)
{
	int fd, ret;

	fd = __sys(open)(dir, O_RDONLY | O_DIRECTORY, 0);
	if (fd < 0)
		__pr_perror("Can't open directory %s: %d", dir, fd);

	ret = __sys(umount2)(dir, MNT_DETACH);
	if (ret) {
		__pr_perror("Can't detach mount %s: %d", dir, ret);
		goto err_close;
	}

	ret = __sys(rmdir)(dir);
	if (ret) {
		__pr_perror("Can't remove tmp dir %s: %d", dir, ret);
		goto err_close;
	}

	return fd;

err_close:
	if (fd >= 0)
		__sys(close)(fd);
	return -1;
}

#define KUP_BUF_SIZE  256

static int int_to_string(unsigned number, char *buf, size_t total) {
	unsigned remainder, quotient, i, len;

	quotient = number;
	len = 0;
	do {
		quotient /= 10;
		len += 1;
	} while (quotient > 0);

	if (len > total - 1)
		return -1;

	quotient = number;
	i = 1;
	do {
		remainder = quotient % 10;
		quotient = quotient / 10;
		buf[len-i] = '0' + remainder;
		i++;
	} while (quotient > 0);
	buf[len] = '\0';

	return len == 0 ? -1 : len;
}

int notifier_kup(enum KUP_HOOK_POINT action, enum nvwa_cmd cmd, bool enable)
{
	int fd, count = 0, retval = 0;
	char buf[KUP_BUF_SIZE] = {0};

	if (!enable)
		return 0;

	fd = __sys(open)(NOTIFY_PROC_PATH, O_WRONLY, 0);
	if (fd == -EACCES) {
		/* there is no priviledge to open file, ignore this condition. */
		pr_info("%s: open %s failed, retval: %d (-EACCES)\n",
				__func__, NOTIFY_PROC_PATH, -EACCES);
		return 0;
	} else if (fd < 0) {
		__pr_perror("%s: Can't open %s: %d\n", __func__, NOTIFY_PROC_PATH, fd);
		return fd;
	}

	retval = int_to_string(action, buf, sizeof(buf)-count);
	if (retval <= 0) {
		__pr_perror("%s: int_to_string error\n", __func__);
		goto err_close;
	}

	buf[retval] = ':';
	count = retval + 1;

	retval = int_to_string(cmd, buf+count, sizeof(buf)-count);
	if (retval <= 0) {
		__pr_perror("%s: int_to_string error\n", __func__);
		goto err_close;
	}

	count += retval;
	retval = __sys(write)(fd, buf, count);
	if (retval < 0)
		__pr_perror("%s: Can't write to %s\n", __func__, NOTIFY_PROC_PATH);

err_close:
	__sys(close)(fd);

	return retval < 0 ? -1 : 0;
}

void do_notifier_rollback(bool rollback, enum notifier_state status)
{
	if (!rollback)
		return;

	switch (status) {
	case POST_UPDATE_KERNEL_COMPLETE:
		notifier_kup(POST_UPDATE_KERNEL, ROLLBACK, true);
		break;
	case UNFREEZE_TO_RUN_COMPLETE:
		notifier_kup(UNFREEZE_TO_RUN, ROLLBACK, true);
		break;
	default:
		break;
	}
}
