#include <fcntl.h>
#include <stdbool.h>
#include <sys/ioctl.h>

#include "cr_options.h"
#include "pstree.h"
#include "mem.h"
#include "vma.h"
#include "pin-mem.h"

bool should_pin_vmae(VmaEntry *vmae)
{
	/*
	 * vDSO area must be always dumped because on restore
	 * we might need to generate a proxy.
	 */
	if (vma_entry_is(vmae, VMA_AREA_VDSO))
		return false;
	/*
	 * In turn VVAR area is special and referenced from
	 * vDSO area by IP addressing (at least on x86) thus
	 * never ever dump its content but always use one provided
	 * by the kernel on restore, ie runtime VVAR area must
	 * be remapped into proper place..
	 */
	if (vma_entry_is(vmae, VMA_AREA_VVAR))
		return false;

	if (vma_entry_is(vmae, VMA_AREA_AIORING))
		return false;
	if (vma_entry_is(vmae, VMA_ANON_PRIVATE))
		return true;

	if (opts.exec_pin_start && vma_entry_is(vmae, VMA_FILE_PRIVATE))
		return true;

	return false;
}

static int pin_one_pmas(int fd, unsigned long start,
			unsigned long *pend, struct pstree_item *item)
{
	int ret;
	unsigned int index = 0;
	unsigned long end;
	unsigned long next = start;
	struct pin_mem_area_set pmas;
	struct _pin_mem_area *pma;

	end = *pend;
	while (start < end) {
		next = (start + ONCE_PIN_MEM_SIZE_LIMIT > end)
			? end : (start + ONCE_PIN_MEM_SIZE_LIMIT);
		pma = &(pmas.mem_area[index]);
		pma->virt_start = start;
		pma->virt_end = next;
		index++;
		start += ONCE_PIN_MEM_SIZE_LIMIT;
		if (index >= MAX_PIN_MEM_AREA_NUM)
			break;
	}

	*pend = next;
	pmas.area_num = index;
	pmas.pid = vpid(item);

	ret = ioctl(fd, SET_PIN_MEM_AREA, &pmas);
	if (ret < 0)
		pr_err("pin mem fail, errno: %s\n", strerror(errno));
	return ret;
}

int pin_vmae(VmaEntry *vmae, struct pstree_item *item)
{
	int fd;
	int ret = 0;
	unsigned long start, end;

	fd = open(PIN_MEM_FILE, O_RDWR);
	if (fd < 0) {
		pr_err("open file: %s fail.\n", PIN_MEM_FILE);
		return -1;
	}

	start = vmae->start;
	while (start < vmae->end) {
		end = vmae->end;
		ret = pin_one_pmas(fd, start, &end, item);
		if (ret < 0)
			break;
		start = end;
	}
	close(fd);
	return ret;
}

int dump_task_special_pages(int pid)
{
	int fd, ret;

	fd = open(PIN_MEM_FILE, O_RDWR, 0);
	if (fd < 0) {
		pr_warn("error open file: %s\n", PIN_MEM_FILE);
		return -1;
	}

	ret = ioctl(fd, DUMP_SPECIAL_PAGES, (unsigned long) &pid);
	if (ret < 0)
		pr_warn("No need DUMP_SPECIAL_PAGES for %d\n", pid);

	close(fd);
	return ret;
}

int restore_task_special_pages(int pid)
{
	int fd, ret;

	fd = open(PIN_MEM_FILE, O_RDWR, 0);
	if (fd < 0) {
		pr_warn("error open file: %s\n", PIN_MEM_FILE);
		return -1;
	}

	ret = ioctl(fd, RETORE_SPECIAL_PAGES, (unsigned long) &pid);
	if (ret < 0)
		pr_warn("No need RETORE_SPECIAL_PAGES for %d\n", pid);

	close(fd);
	return ret;
}

int clear_pin_mem(int pid)
{
	int fd, ret;

	fd = open(PIN_MEM_FILE, O_RDWR, 0);
	if (fd < 0) {
		pr_warn("error open file: %s\n", PIN_MEM_FILE);
		return -1;
	}

	ret = ioctl(fd, CLEAR_PIN_MEM_AREA, (unsigned long) &pid);
	if (ret < 0) {
		pr_warn("clear pin mem fail: %d\n", pid);
	}

	close(fd);
	return ret;
}

int init_pagemap_read(int para)
{
	int fd, ret;

	fd = open(PIN_MEM_FILE, O_RDWR, 0);
	if (fd < 0) {
		pr_warn("error open file: %s\n", PIN_MEM_FILE);
		return -1;
	}

	ret = ioctl(fd, INIT_PAGEMAP_READ, (unsigned long) &para);
	if (ret < 0) {
		pr_warn("Init pagemap read fail, errno: %s\n", strerror(errno));
	}

	close(fd);
	return ret;
}

