/*
 * Target:
 *   parallel dump process
 */

#include <string.h>
#include <errno.h>
#include <pthread.h>
#include <sys/sysinfo.h>

#include "pstree.h"
#include "log.h"
#include "taskqueue.h"

/*
 * Sometimes, only one cpu can be used which is bad for parallel routine.
 * Therefore, set cpu affinity for criu routine.
 */
static int set_cpuaffinity(void)
{
	cpu_set_t *set;
	int num_cpus = get_nprocs_conf();
	size_t cpusetsize = CPU_ALLOC_SIZE(num_cpus);
	int retval;

	set = CPU_ALLOC(num_cpus);
	memset(set, 0xff, cpusetsize);

	retval = sched_setaffinity(getpid(), cpusetsize, set);
	if (retval != 0)
		pr_err("sched_setaffinity failed: %s\n", strerror(errno));

	CPU_FREE(set);

	return retval;
}

int init_parallel_env(void)
{
	return set_cpuaffinity();
}

static void *collect_mappings_routine(void *_arg)
{
	struct mappings_info *info = _arg;

	info->queue_result = collect_mappings(info->pid, info->vmas, info->dump_file);
	return NULL;
}

int dump_filemap(struct vma_area *vma_area, int fd); /* defined in criu/cr-dump.c */

int start_collect_mappings_thread(void)
{
	struct pstree_item *pi;
	struct mappings_info *info;

	for_each_pstree_item(pi) {
		/* disable parallel collect for non-root item because of the
		 * concurrence.
		 */
		if (pi->pid->real != root_item->pid->real)
			continue;

		info = &pi->maps_info;

		info->vmas = xmalloc(sizeof(struct vm_area_list));
		if (info->vmas == NULL) {
			pr_err("xzalloc vmas no memory\n");
			return -1;
		}
		vm_area_list_init(info->vmas);

		info->pid = pi->pid->real;
		info->dump_file = dump_filemap;
		info->queue_routine = collect_mappings_routine;
		info->queue_arg = info;

		pr_info("Start thread to collect %d mappings\n", info->pid);

		if (taskqueue_create(&info->queue) < 0) {
			pr_err("parallel_collect_mappings failed: %s\n", strerror(errno));
			free(info->vmas);
			/*
			 * Don't care other threads status, use `exit_group()`
			 * to ensure all threads exit.
			 */
			return -1;
		}
	}

	return 0;
}

int end_collect_mappings_thread(struct pstree_item *item)
{
	struct mappings_info *info = &item->maps_info;
	int retval;

	/* disable parallel collect for non-root item because of the
	 * concurrence.
	 */
	if (root_item->pid->real != item->pid->real)
		return 0;

	retval = taskqueue_join(&info->queue);
	if (retval != 0 || info->queue_result != 0) {
		pr_err("taskqueue_join failed, retval %d(errno %d: %s),"
			       " queue_result: %d\n",
			retval,
			retval == 0 ? 0 : errno,
			retval == 0 ? "nil" : strerror(errno),
			info->queue_result);
		retval = -1;
	}

	pr_info("End thread to collect %d mappings\n", info->pid);

	/*
	 * Don't care other threads status, use `exit_group()` to ensure all
	 * threads exit.
	 */
	return retval;
}
